import inspect
import operator
from itertools import chain
from os import environ
from typing import Generator, Iterable, TypeVar, Union

import numpy as np
import pandas as pd
from joblib import Parallel, cpu_count, delayed
from mne.utils import config, logger
from scipy.signal import get_window


def iterate_axes(arr: np.ndarray, axes: tuple[int, ...], index=(), axis=0):
    """Iterate over all possible indices for a set of axes

    Parameters
    ----------
    arr : np.ndarray
        The array to iterate over
    axes : tuple[int]
        The axes to iterate over
    index : tuple[int]
        The current index
    axis : int
        The current axis

    Yields
    ------
    tuple[slice]
        The indices for the current iteration

    Examples
    --------
    >>> arr = np.arange(24).reshape(2, 3, 4)
    >>> for sl in iterate_axes(arr, (0, 1)):
    ...     print(arr[sl])
    [0 1 2 3]
    [4 5 6 7]
    [ 8  9 10 11]
    [12 13 14 15]
    [16 17 18 19]
    [20 21 22 23]
    """
    if axis < len(axes):
        for i in range(arr.shape[axes[axis]]):
            yield from iterate_axes(arr, axes, index + (i,), axis + 1)
    else:
        # Create a tuple of slices for all axes
        slices = [slice(None)] * arr.ndim
        for axis, i in zip(axes, index):
            slices[axis] = i
        yield tuple(slices)


def ensure_int(x, name: str = 'unknown', must_be: str = 'an int', *, extra=''):
    """Ensure a variable is an integer.

    Parameters
    ----------
    x : object
        The object to check.
    name : str
        The name of the variable to check.
    must_be : str
        The type of the variable to check.
    extra : str
        Extra text to add to the error message.

    Notes
    -----
    This is preferred over numbers.Integral, see:
    https://github.com/scipy/scipy/pull/7351#issuecomment-299713159

    Examples
    --------
    >>> ensure_int(1)
    1
    >>> ensure_int(1.0)
    Traceback (most recent call last):
    ...
    TypeError: unknown must be an int, got <class 'float'>
    >>> ensure_int('1')
    Traceback (most recent call last):
    ...
    TypeError: unknown must be an int, got <class 'str'>
    >>> ensure_int('1.0', extra='a string')
    Traceback (most recent call last):
    ...
    TypeError: unknown must be an int a string, got <class 'str'>
    """

    extra = f' {extra}' if extra else extra
    try:
        # someone passing True/False is much more likely to be an error than
        # intentional usage
        if isinstance(x, bool):
            raise TypeError()
        x = int(operator.index(x))
    except TypeError:
        raise TypeError(f'{name} must be {must_be}{extra}, got {type(x)}')
    return x


def validate_type(item, types):
    """Validate the type of an object.

    Parameters
    ----------
    item : object
        The object to check.
    types : type
        The type to check against.
    """
    try:
        if isinstance(types, TypeVar):
            check = isinstance(item, types.__constraints__)
        elif types is int:
            ensure_int(item)
            check = True
        elif types is float:
            check = is_number(item)
        else:
            check = isinstance(item, types)
    except TypeError:
        check = False
    if not check:
        raise TypeError(
            f"must be an instance of {types}, "
            f"got {type(item)} instead.")


def is_number(s) -> bool:
    """Check if an object is a number

    Parameters
    ----------
    s : object
        The object to check

    Returns
    -------
    bool
        True if the object is a number, False otherwise
    """
    if isinstance(s, str):
        try:
            float(s)
            return True
        except ValueError:
            return False
    elif isinstance(s, (np.number, int, float)):
        return True
    elif isinstance(s, pd.DataFrame):
        try:
            s.astype(float)
            return True
        except Exception:
            return False
    elif isinstance(s, pd.Series):
        try:
            pd.to_numeric(s)
            return True
        except Exception:
            return False
    else:
        return False


def proc_array(func: callable, arr_in: np.ndarray, axes: int | tuple[int] = 0,
               n_jobs: int = None, desc: str = "Slices", inplace: bool = True,
               **kwargs) -> np.ndarray:
    """Execute a function in parallel over slices of an array

    Parameters
    ----------
    func : callable
        The function to execute
    arr_in : np.ndarray
        The array to slice
    axes : int | tuple[int]
        The axes to slice over
    n_jobs : int
        The number of jobs to run in parallel
    desc : str
        The description to use for the progress bar
    inplace : bool
        Whether to modify the input array in place

    Returns
    -------
    np.ndarray
        The output of the function, same shape as the input array

    Examples
    --------
    >>> def square(x):
    ...     return x ** 2
    >>> proc_array(square, np.arange(10))
    array([ 0,  1,  4,  9, 16, 25, 36, 49, 64, 81])
    """

    if isinstance(axes, int):
        axes = (axes,)

    if inplace:
        arr_out = arr_in
    else:
        arr_out = arr_in.copy()

    # Get the cross-section indices and array input generator
    cross_sect_ind = list(np.ndindex(*[arr_in.shape[axis] for axis in axes]))
    array_gen = list(arr_in[indices] for indices in cross_sect_ind)

    gen = Parallel(n_jobs, return_as='generator', verbose=40)(
        delayed(func)(x_, **kwargs) for x_ in array_gen)

    # Create process pool and apply the function in parallel
    for out, ind in zip(gen, cross_sect_ind):
        arr_out[ind] = out

    return arr_out


def parallelize(func: callable, ins: Iterable, verbose: int = 10,
                n_jobs: int = None, **kwargs) -> list | None:
    """Parallelize a function to run on multiple processors.

    This function is a wrapper for joblib.Parallel. It will automatically
    determine the number of jobs to run in parallel based on the number of
    cores available on the system. It will also automatically set the
    temp_folder and max_nbytes parameters for joblib.Parallel based on the
    MNE_CACHE_DIR and MNE_MEMMAP_MIN_SIZE parameters in mne-python's
    configuration file.

    Notes
    -----
    If the elements of the par_var iterable are tuples, the function will be
    called with the tuple unpacked, setting each item in the tuple to be
    assigned to a separate argument. If the elements of the par_var iterable
    are not tuples, the function will be called with the element as the first
    argument.


    Parameters
    ----------
    func : callable
        The function to parallelize
    ins : Iterable
        The iterable to parallelize over
    n_jobs : int
        The number of jobs to run in parallel. If None, will use all
        available cores. If -1, will use all available cores.
    **kwargs
        Additional keyword arguments to pass to the function

    Returns
    -------
    list
        The output of the function for each element in par_var

    Examples
    --------
    >>> def square(x):
    ...     return x ** 2
    >>> parallelize(square, [1, 2, 3])
    [1, 4, 9]
    """

    assert ins
    if n_jobs is None:
        if 'n_jobs' in inspect.getfullargspec(func).args:
            kwargs['n_jobs'] = -2
        else:
            n_jobs = -2
    settings = dict(verbose=verbose)
    settings['prefer'] = kwargs.pop('prefer', None)
    settings['backend'] = kwargs.pop('backend', None)
    settings['mmap_mode'] = kwargs.pop('mmap_mode', 'r')
    settings['require'] = kwargs.pop('require', None)

    env = dict(**environ)
    if config.get_config('MNE_CACHE_DIR') is not None:
        settings['temp_folder'] = config.get_config('MNE_CACHE_DIR')
    elif 'TEMP' in env.keys():
        settings['temp_folder'] = env['TEMP']
    else:
        settings['temp_folder'] = None

    if config.get_config('MNE_MEMMAP_MIN_SIZE') is not None:
        settings['max_nbytes'] = config.get_config('MNE_MEMMAP_MIN_SIZE')
    else:
        settings['max_nbytes'] = get_mem()

    for var in ins:
        if isinstance(var, tuple):
            x_is_tup = True
        elif isinstance(ins, Generator):
            x_is_tup = False
            ins = chain((var,), ins)
        else:
            x_is_tup = False
        break

    if x_is_tup:
        return Parallel(n_jobs, **settings)(delayed(func)(
            *x_, **kwargs) for x_ in ins)
    else:
        return Parallel(n_jobs, **settings)(delayed(func)(
            x_, **kwargs) for x_ in ins)


def get_mem() -> int:
    """Get the amount of memory to use for parallelization.

    Returns
    -------
    float | int
        The amount of memory to use for parallelization
    """
    from psutil import virtual_memory
    ram_per = virtual_memory().total >> 0 // cpu_count()
    return ram_per


def sliding_window(x_data: np.ndarray, labels: np.ndarray,
                   scorer: callable, window_size: int = 20, axis: int = -1,
                   n_jobs: int = -3, **kwargs) -> np.ndarray:
    """Compute a function over a sliding window.

    Parameters
    ----------
    x_data : np.ndarray, shape (..., trials, time)
        The data to compute the function over
    labels : np.ndarray, shape (trials,)
        The labels for each trial
    scorer : callable
        The function to compute over the sliding window. Must take two
        arguments, the data and the labels.
    window_size : int
        The size of the sliding window
    axis : int
        The axis to compute the sliding window over
    n_jobs : int
        The number of jobs to run in parallel

    Returns
    -------
    np.ndarray
        The output of the function, shape (..., time - window_size + 1)

    Examples
    --------
    >>> def square(x, labels):
    ...     return np.mean(x ** 2, where=labels == 1)
    >>> x_data = np.arange(40).reshape(4, 10)
    >>> labels = np.array([0, 1, 1])
    >>> sliding_window(x_data, labels, square, window_size=3)
    array([397.5, 431.5, 467.5, 505.5, 545.5, 587.5, 631.5])
    """

    # make windowing generator
    axis = x_data.ndim + axis if axis < 0 else axis
    slices = (slice(start, start + window_size)
              for start in range(0, x_data.shape[axis] - window_size))
    idxs = (tuple(slice(None) if i != axis else sl for i in
                  range(x_data.ndim)) for sl in slices)

    # Use joblib to parallelize the computation
    gen = Parallel(n_jobs=n_jobs, return_as='generator', verbose=40)(
        delayed(scorer)(x_data[idx], labels, **kwargs) for idx in idxs)

    # initialize output array by running 1 job and get the shape
    mat = next(gen)
    out = np.zeros((x_data.shape[axis] - window_size, *mat.shape),
                   dtype=mat.dtype)
    out[0] = mat

    # fill in the rest of the output array
    for i, mat in enumerate(gen):
        out[i + 1] = mat

    return out


###############################################################################
# Constant overlap-add processing class


def _check_store(store):
    if isinstance(store, np.ndarray):
        store = [store]
    if isinstance(store, (list, tuple)) and all(isinstance(s, np.ndarray)
                                                for s in store):
        store = _Storer(*store)
    if not callable(store):
        raise TypeError('store must be callable, got type %s'
                        % (type(store),))
    return store


class COLA:
    """Constant overlap-add processing helper.

    Parameters
    ----------
    process : callable
        A function that takes a chunk of input data with shape
        `(n_channels, n_samples)` and processes it.
    store : callable | ndarray
        A function that takes a completed chunk of output data.
        Can also be an `ndarray`, in which case it is treated as the
        output data in which to store the results.
    n_total : int
        The total number of samples.
    n_samples : int
        The number of samples per window.
    n_overlap : int
        The overlap between windows.
    window : str
        The window to use. Default is "hann".
    tol : float
        The tolerance for COLA checking.
    n_jobs : int
        The number of jobs to run in parallel.
    verbose : bool
        If True, print a message when the COLA condition is not met.

    Notes
    -----
    This will process data using overlapping windows to achieve a constant
    output value. For example, for ``n_total=27``, ``n_samples=10``,
    ``n_overlap=5`` and ``window='triang'``
    ::

        1 _____               _______
          |    \\   /\\   /\\   /
          |     \\ /  \\ /  \\ /
          |      x    x    x
          |     / \\  / \\  / \\
          |    /   \\/   \\/   \\
        0 +----|----|----|----|----|-
          0    5   10   15   20   25

    This produces four windows: the first three are the requested length
    (10 samples) and the last one is longer (12 samples). The first and last
    window are asymmetric.
    """

    def __init__(self, process, store, n_total, n_samples, n_overlap, sfreq,
                 window='hann', tol=1e-10, *, verbose=None):
        n_samples = ensure_int(n_samples, 'n_samples')
        n_overlap = ensure_int(n_overlap, 'n_overlap')
        n_total = ensure_int(n_total, 'n_total')
        if n_samples <= 0:
            raise ValueError('n_samples must be > 0, got %s' % (n_samples,))
        if n_overlap < 0:
            raise ValueError('n_overlap must be >= 0, got %s' % (n_overlap,))
        if n_total < 0:
            raise ValueError('n_total must be >= 0, got %s' % (n_total,))
        self._n_samples = int(n_samples)
        self._n_overlap = int(n_overlap)
        del n_samples, n_overlap
        if n_total < self._n_samples:
            raise ValueError('Number of samples per window (%d) must be at '
                             'most the total number of samples (%s)'
                             % (self._n_samples, n_total))
        if not callable(process):
            raise TypeError('process must be callable, got type %s'
                            % (type(process),))
        self._process = process
        self._step = self._n_samples - self._n_overlap
        self._store = _check_store(store)
        self._idx = 0
        self._in_buffers = self._out_buffers = None

        # Create our window boundaries
        window_name = window if isinstance(window, str) else 'custom'
        self._window = get_window(window, self._n_samples,
                                  fftbins=(self._n_samples - 1) % 2)
        self._window /= _check_cola(self._window, self._n_samples, self._step,
                                    window_name, tol=tol)
        self.starts = np.arange(0, n_total - self._n_samples + 1, self._step)
        self.stops = self.starts + self._n_samples
        delta = n_total - self.stops[-1]
        self.stops[-1] = n_total
        sfreq = float(sfreq)
        pl = 's' if len(self.starts) != 1 else ''
        if verbose:
            logger.info('    Processing %4d data chunk%s of (at least) %0.1f '
                        'sec with %0.1f sec overlap and %s windowing'
                        % (len(self.starts), pl, self._n_samples / sfreq,
                           self._n_overlap / sfreq, window_name))
        del window, window_name
        if delta > 0 and verbose:
            logger.info('    The final %0.3f sec will be lumped into the '
                        'final window' % (delta / sfreq,))

    @property
    def _in_offset(self):
        """Compute from current processing window start and buffer len."""
        return self.starts[self._idx] + self._in_buffers[0].shape[-1]

    def feed(self, *datas, verbose=None, **kwargs):
        """Pass in a chunk of data."""
        # Append to our input buffer
        if self._in_buffers is None:
            self._in_buffers = [None] * len(datas)
        if len(datas) != len(self._in_buffers):
            raise ValueError('Got %d array(s), needed %d'
                             % (len(datas), len(self._in_buffers)))
        for di, data in enumerate(datas):
            if not isinstance(data, np.ndarray) or data.ndim < 1:
                raise TypeError('data entry %d must be an 2D ndarray, got %s'
                                % (di, type(data),))
            if self._in_buffers[di] is None:
                # In practice, users can give large chunks, so we use
                # dynamic allocation of the in buffer. We could save some
                # memory allocation by only ever processing max_len at once,
                # but this would increase code complexity.
                self._in_buffers[di] = np.empty(
                    data.shape[:-1] + (0,), data.dtype)
            if data.shape[:-1] != self._in_buffers[di].shape[:-1] or \
                    self._in_buffers[di].dtype != data.dtype:
                raise TypeError('data must dtype %s and shape[:-1]==%s, '
                                'got dtype %s shape[:-1]=%s'
                                % (self._in_buffers[di].dtype,
                                   self._in_buffers[di].shape[:-1],
                                   data.dtype, data.shape[:-1]))
            # logger.debug('    + Appending %d->%d'
            #              % (self._in_offset, self._in_offset + data.shape[
            #              -1]))
            self._in_buffers[di] = np.concatenate(
                [self._in_buffers[di], data], -1)
            if self._in_offset > self.stops[-1]:
                raise ValueError('data (shape %s) exceeded expected total '
                                 'buffer size (%s > %s)'
                                 % (data.shape, self._in_offset,
                                    self.stops[-1]))
        # preallocate data to chunks
        data_chunks = map(lambda x, y: data[x:y], self.starts, self.stops)
        out_chunks = map(lambda d: self._process(d, **kwargs), data_chunks)

        # overlap add to buffer
        while self._idx < len(self.starts) and \
                self._in_offset >= self.stops[self._idx]:
            start, stop = self.starts[self._idx], self.stops[self._idx]
            this_len = stop - start
            this_window = self._window.copy()
            if self._idx == len(self.starts) - 1:
                this_window = np.pad(
                    self._window, (0, this_len - len(this_window)), 'constant')
                for offset in range(self._step, len(this_window), self._step):
                    n_use = len(this_window) - offset
                    this_window[offset:] += self._window[:n_use]
            if self._idx == 0:
                for offset in range(self._n_samples - self._step, 0,
                                    -self._step):
                    this_window[:offset] += self._window[-offset:]
            # logger.debug('    * Processing %d->%d' % (start, stop))
            this_proc = [in_[..., :this_len].copy()
                         for in_ in self._in_buffers]
            if not all(proc.shape[-1] == this_len == this_window.size
                       for proc in this_proc):
                raise RuntimeError('internal indexing error')
            outs = next(out_chunks)
            if self._out_buffers is None:
                max_len = np.max(self.stops - self.starts)
                self._out_buffers = [np.zeros(o.shape[:-1] + (max_len,),
                                              o.dtype) for o in outs]
            for oi, out in enumerate(outs):
                out *= this_window
                self._out_buffers[oi][..., :stop - start] += out
            self._idx += 1
            if self._idx < len(self.starts):
                next_start = self.starts[self._idx]
            else:
                next_start = self.stops[-1]
            delta = next_start - self.starts[self._idx - 1]
            for di in range(len(self._in_buffers)):
                self._in_buffers[di] = self._in_buffers[di][..., delta:]
            # logger.debug('    - Shifting input/output buffers by %d samples'
            #              % (delta,))
            self._store(*[o[..., :delta] for o in self._out_buffers])
            for ob in self._out_buffers:
                ob[..., :-delta] = ob[..., delta:]
                ob[..., -delta:] = 0.


def _check_cola(win, nperseg, step, window_name, tol=1e-10):
    """Check whether the Constant OverLap Add (COLA) constraint is met."""
    # adapted from SciPy
    binsums = np.sum([win[ii * step:(ii + 1) * step]
                      for ii in range(nperseg // step)], axis=0)
    if nperseg % step != 0:
        binsums[:nperseg % step] += win[-(nperseg % step):]
    const = np.median(binsums)
    deviation = np.max(np.abs(binsums - const))
    if deviation > tol:
        raise ValueError('segment length %d with step %d for %s window '
                         'type does not provide a constant output '
                         '(%g%% deviation)'
                         % (nperseg, step, window_name,
                            100 * deviation / const))
    return const


class _Storer(object):
    """Store data in chunks."""

    def __init__(self, *outs, picks=None):
        for oi, out in enumerate(outs):
            if not isinstance(out, np.ndarray) or out.ndim < 1:
                raise TypeError('outs[oi] must be >= 1D ndarray, got %s'
                                % (out,))
        self.outs = outs
        self.idx = 0
        self.picks = picks

    def __call__(self, *outs):
        if (len(outs) != len(self.outs) or
                not all(out.shape[-1] == outs[0].shape[-1] for out in outs)):
            raise ValueError('Bad outs')
        idx = (Ellipsis,)
        if self.picks is not None:
            idx += (self.picks,)
        stop = self.idx + outs[0].shape[-1]
        idx += (slice(self.idx, stop),)
        for o1, o2 in zip(self.outs, outs):
            o1[idx] = o2
        self.idx = stop



from setuptools.extension import Extension
import numpy as np
from setuptools import setup, find_packages
import os.path as op
from os import getcwd
import sys
import glob


_numpy_abs = np.get_include()  # get the numpy include path

def get_file_list(path, ext):
    all_files = glob.glob(op.join(path, f"*{ext}"))
    for file in all_files:
        elem = op.split(file)
        name = op.splitext(".".join(elem))[0]
        yield name, [file]


npymath_path = op.normpath(op.join(_numpy_abs, '..', 'lib'))
npyrandom_path = op.normpath(op.join(_numpy_abs, '..', '..', 'random', 'lib'))
lib_path = [npymath_path, npyrandom_path]
if sys.platform == 'win32':
    compile_args = ["/O2"]
elif sys.platform == 'linux':
    compile_args = ["-O3"]
else:
    raise NotImplementedError(f"Platform {sys.platform} not supported.")

try:
    _numpy_abs = op.relpath(_numpy_abs, getcwd())
except ValueError:
    pass


kwargs = dict(include_dirs=[_numpy_abs],
              # includes for numpy
              library_dirs=lib_path,  # libraries to link
              libraries=["npyrandom", "npymath"],  # math library
              extra_compile_args=compile_args,  # compile optimization flag
              language="c",  # can be "c" or "c++"
              define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
              )

try:
    from Cython.Build import cythonize
    if not op.exists("ieeg/timefreq/hilbert.pyx"):
        raise ImportError("Cython file not found.")
    extensions = [
        Extension(
            "ieeg.calc._fast.*",  # the module name exposed to python
            ['ieeg/calc/_fast/*.pyx'],  # the Cython source file
            **kwargs
        ),
        Extension(
            "ieeg.timefreq.hilbert",  # the module name exposed to python
            ['ieeg/timefreq/hilbert.pyx'],  # the Cython source file
            **kwargs
        )]
    extensions = cythonize(extensions)
except ImportError:
    USE_CYTHON = False
    print("Cython not found. Using C files.")
    if not op.exists("ieeg/timefreq/hilbert.c"):
        ValueError("C file not found.")

extensions = [
    Extension(name, source, **kwargs) for (name, source) in
    get_file_list("ieeg/calc/_fast", ".c")]
extensions += [
    Extension(
        "ieeg.timefreq.hilbert",  # the module name exposed to python
        ['ieeg/timefreq/hilbert.c'],  # the Cython source file
        **kwargs
    ),
]

setup(
    name='ieeg',
    version='0.3',
    packages=find_packages(
        where='.',
        include=['ieeg*'],
    ),
    package_dir={"": "."},
    description='A Python package for iEEG data processing.',
    author='Aaron Earle-Richardson',
    author_email='ae166@duke.edu',
    url='https://github.com/coganlab/IEEG_Pipelines',
    ext_modules=extensions,
)


from typing import Union

import numpy as np
from mne.epochs import BaseEpochs
from mne.evoked import Evoked
from mne.io import base
from mne.time_frequency import EpochsTFR, tfr
from mne.utils import fill_doc, verbose

from ieeg import Signal
from ieeg.process import ensure_int, parallelize, validate_type


def to_samples(time_length: Union[str, int], sfreq: float) -> int:
    """Convert a time length to a number of samples.

    Parameters
    ----------
    time_length : str | int
        The time length to convert. If a string, it must be a human-readable
        time, e.g. "10s".
    sfreq : float
        The sampling frequency.

    Returns
    -------
    samples : int
        The number of samples.
    """
    validate_type(time_length, (str, int))
    if isinstance(time_length, str):
        time_length = time_length.lower()
        err_msg = ('filter_length, if a string, must be a '
                   'human-readable time, e.g. "0.7s", or "700ms", not '
                   '"%s"' % time_length)
        low = time_length.lower()
        if low.endswith('us'):
            mult_fact = 1e-6
            time_length = time_length[:-2]
        elif low.endswith('ms'):
            mult_fact = 1e-3
            time_length = time_length[:-2]
        elif low[-1] == 's':
            mult_fact = 1
            time_length = time_length[:-1]
        elif low.endswith('sec'):
            mult_fact = 1
            time_length = time_length[:-3]
        elif low[-1] == 'm':
            mult_fact = 60
            time_length = time_length[:-1]
        elif low.endswith('min'):
            mult_fact = 60
            time_length = time_length[:-3]
        else:
            raise ValueError(err_msg)
        # now get the number
        try:
            time_length = float(time_length)
        except ValueError:
            raise ValueError(err_msg)
        time_length = max(int(np.ceil(time_length * mult_fact * sfreq)), 1)
    time_length = ensure_int(time_length, 'filter_length')
    return time_length


@fill_doc
def crop_pad(inst: Signal, pad: str, copy: bool = False) -> Signal:
    """Crop and pad an instance.

    Parameters
    ----------
    inst : instance of Raw, Epochs, or Evoked
        The instance to crop and pad.
    pad : str
        The amount of time to pad the instance. If a string, it must be a
        human-readable time, e.g. "10s".
    copy : bool, optional
        If True, a copy of x, filtered, is returned. Otherwise, it operates
        on x in place. Defaults to False.

    Returns
    -------
    inst : instance of Raw, Epochs, or Evoked
        The cropped and de-padded instance.
    """
    if copy:
        out = inst.copy()
    else:
        out = inst
    pad = to_samples(pad, inst.info['sfreq']) / inst.info['sfreq']
    out.crop(tmin=inst.tmin + pad, tmax=inst.tmax - pad)
    return out


@verbose
def wavelet_scaleogram(inst: BaseEpochs, f_low: float = 2,
                       f_high: float = 1000, k0: int = 6, n_jobs: int = 1,
                       decim: int = 1, verbose=10) -> EpochsTFR:
    """Compute the wavelet scaleogram.



    Parameters
    ----------
    inst : instance of Raw, Epochs, or Evoked
        The instance to compute the wavelet scaleogram for.
    f_low : float
        The lowest frequency to compute the scaleogram for.
    f_high : float
        The highest frequency to compute the scaleogram for.
    k0 : int
        The wavelet parameter.
    n_jobs : int
        The number of jobs to run in parallel.
    decim : int
        The decimation factor.
    verbose : int
        The verbosity level.

    Returns
    -------
    scaleogram : instance of EpochsTFR
        The wavelet scaleogram.

    Notes
    -----
    Similar to https://www.mathworks.com/help/wavelet/ref/cwt.html

    Examples
    --------
    >>> import mne
    >>> from ieeg.io import raw_from_layout
    >>> from ieeg.navigate import trial_ieeg
    >>> from bids import BIDSLayout
    >>> with mne.use_log_level(0):
    ...     bids_root = mne.datasets.epilepsy_ecog.data_path()
    ...     layout = BIDSLayout(bids_root)
    ...     raw = raw_from_layout(layout, subject="pt1", preload=True,
    ...     extension=".vhdr", verbose=False)
    ...     epochs = trial_ieeg(raw, ['AST1,3', 'G16'], (-1, 2), verbose=False)
    >>> wavelet_scaleogram(epochs, n_jobs=-2, decim=10) # doctest: +ELLIPSIS
    Using data from preloaded Raw for 2 events and 3001 original time points...
        Getting epoch for 85800-88801
        Getting epoch for 90760-93761
    0 bad epochs dropped
    Data is self data: False
    <TFR from Epochs, unknown method | 2 epochs × 98 channels × 46 freqs × ...

    """
    data = inst.get_data(copy=False)

    f = np.fft.fft(data - np.mean(data, axis=-1, keepdims=True))

    daughter, period = calculate_wavelets(inst.info['sfreq'], f_high, f_low,
                                          data.shape[-1], k0)

    wave = np.empty((f.shape[0], f.shape[1], len(period),
                     data[..., ::decim].shape[-1]), dtype=np.float64)
    # ch X trials X freq X time
    ins = ((f[:, None, i], i) for i in range(f.shape[1]))

    def _ifft_abs(x, i):
        np.abs(np.fft.ifft(x * np.tile(daughter, (
            f.shape[0], 1, 1)))[..., ::decim], out=wave[:, i])

    parallelize(_ifft_abs, ins, require='sharedmem', n_jobs=n_jobs,
                verbose=verbose)

    return EpochsTFR(inst.info, wave, inst.times[::decim], 1 / period)


def calculate_wavelets(sfreq: float, f_high: float, f_low: float,
                       n_samples: int, k0: int = 6):
    """Calculate Morlet wavelets for a range of frequencies.

    Parameters
    ----------
    sfreq : float
        The sampling frequency.
    f_high : float
        The highest frequency to compute the scaleogram for.
    f_low : float
        The lowest frequency to compute the scaleogram for.
    n_samples : int
        The number of samples.
    k0 : int
        The wavelet parameter. Indicates the center frequency of the wavelet.

    Returns
    -------
    daughter : ndarray
        The wavelets.
    period : ndarray
        The periods.

    Examples
    --------
    >>> daughter, period = calculate_wavelets(1000, 100, 2, 1000)
    >>> daughter.shape
    (29, 1001)
    >>> period.shape
    (29,)
    """

    dt = 1 / sfreq
    s0 = 1 / (f_high + (0.1 * f_high))  # the smallest resolvable scale
    n = n_samples
    J1 = (np.log2(n * dt / s0)) / 0.2  # (J1 determines the largest scale)

    k = np.arange(np.fix(n / 2)) + 1
    k = k * ((2 * np.pi) / (n * dt))
    kr = (-k).tolist()
    kr.reverse()
    k = np.array([0] + k.tolist() + kr)

    scale = s0 * np.power(2., (np.arange(0, J1) * 0.2))
    fourier_factor = (4 * np.pi) / (k0 + np.sqrt(2 + np.square(k0)))
    period = fourier_factor * scale
    xxx = np.min(np.where((1. / period) < f_low))
    period = np.flip(period[:xxx])
    scale = np.flip(scale[:xxx])

    scale1 = scale
    period = fourier_factor * scale1

    expnt = -np.square(scale1[:, None] * k[None, :] - k0) / 2. * (k > 0.)
    norm = np.sqrt(scale1 * k[2]) * (np.power(np.pi, (-0.25))) * np.sqrt(n)
    daughter = norm[:, None] * np.exp(expnt)
    daughter = daughter * (k > 0.)

    return daughter, period


def roundup(x: float) -> int:
    """Round up to the nearest integer."""
    n, d = divmod(x, 1)
    return int(n) + (d > 0)


@verbose
def cwt(inst: BaseEpochs, f_low: float = 2,
        f_high: float = 1000, n_jobs: int = 1, k0: int = 6,
        decim: int = 1, verbose=10) -> EpochsTFR:
    """Compute the wavelet scaleogram.



    Parameters
    ----------
    inst : instance of Raw, Epochs, or Evoked
        The instance to compute the wavelet scaleogram for.
    f_low : float
        The lowest frequency to compute the scaleogram for.
    f_high : float
        The highest frequency to compute the scaleogram for.
    k0 : int
        The wavelet parameter.
    n_jobs : int
        The number of jobs to run in parallel.
    decim : int
        The decimation factor.
    verbose : int
        The verbosity level.

    Returns
    -------
    scaleogram : instance of EpochsTFR
        The wavelet scaleogram.

    Notes
    -----
    Similar to https://www.mathworks.com/help/wavelet/ref/cwt.html
    """
    data = inst.get_data()  # (trials X channels X timepoints)
    data -= np.mean(data, axis=-1, keepdims=True)

    # Ws, period = calculate_wavelets(inst.info['sfreq'], f_high, f_low,
    # data.shape[-1])
    freqs = np.logspace(np.log10(f_low), np.log10(f_high), 50)
    common_factor = (data.shape[-1] + 1) * np.pi / 5 / inst.info['sfreq']
    n_cycles = np.min(freqs) * common_factor
    wavelets = tfr.morlet(inst.info['sfreq'], freqs, n_cycles=n_cycles,
                          zero_mean=True)
    wave = np.empty((data.shape[0], data.shape[1], len(freqs),
                     roundup(data.shape[2] / decim)), dtype=np.float64)

    def _cwt(x, i):
        wave[i] = tfr.cwt(x, wavelets, use_fft=True, decim=decim)

    # perform the cwt across trials and channels
    # iterate through fisrt and second dimension
    ins = ((d, i) for i, d in enumerate(data))
    parallelize(_cwt, ins, require='sharedmem', n_jobs=n_jobs,
                verbose=verbose)

    return EpochsTFR(inst.info, wave, inst.times[::decim], freqs)


def _check_filterable(x: Union[Signal, np.ndarray],
                      kind: str = 'filtered',
                      alternative: str = 'filter') -> np.ndarray:
    # Let's be fairly strict about this -- users can easily coerce to ndarray
    # at their end, and we already should do it internally any time we are
    # using these low-level functions. At the same time, let's
    # help people who might accidentally use low-level functions that they
    # shouldn't use by pushing them in the right direction
    if isinstance(x, (base.BaseRaw, BaseEpochs, Evoked)):
        try:
            name = x.__class__.__name__
        except Exception:
            pass
        else:
            raise TypeError(
                'This low-level function only operates on np.ndarray '
                f'instances. To get a {kind} {name} instance, use a method '
                f'like `inst_new = inst.copy().{alternative}(...)` '
                'instead.')
    validate_type(x, (np.ndarray, list, tuple))
    x = np.asanyarray(x)
    if x.dtype != np.float64:
        raise ValueError('Data to be %s must be real floating, got %s'
                         % (kind, x.dtype,))
    return x


if __name__ == "__main__":
    # Description: Produce spectrograms for each subject
    from ieeg.io import get_data, raw_from_layout
    from ieeg.navigate import trial_ieeg, crop_empty_data, \
        outliers_to_nan
    import os
    import numpy as np

    # check if currently running a slurm job
    HOME = os.path.expanduser("~")

    if 'SLURM_ARRAY_TASK_ID' in os.environ.keys():
        LAB_root = os.path.join(HOME, "workspace", "CoganLab")
        layout = get_data("SentenceRep", root=LAB_root)
        subjects = list(int(os.environ['SLURM_ARRAY_TASK_ID']))
    else:  # if not then set box directory
        LAB_root = os.path.join(HOME, "Box", "CoganLab")
        layout = get_data("SentenceRep", root=LAB_root)
        subjects = layout.get(return_type="id", target="subject")

    for sub in subjects:
        if sub != "D0029":
            continue
        # Load the data
        filt = raw_from_layout(layout.derivatives['clean'], subject=sub,
                               extension='.edf', desc='clean', preload=False)

        # Crop raw data to minimize processing time
        good = crop_empty_data(filt, ).copy()

        # good.info['bads'] = channel_outlier_marker(good, 3, 2)
        good.drop_channels(good.info['bads'])
        good.load_data()

        ch_type = filt.get_channel_types(only_data_chs=True)[0]
        good.set_eeg_reference(ref_channels="average", ch_type=ch_type)

        # Remove intermediates from mem
        good.plot()

        # epoching and trial outlier removal

        save_dir = os.path.join(layout.root, 'derivatives', 'spec',
                                'wavelet_test', sub)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        for epoch, t, name in zip(
                ("Start", "Word/Response/LS", "Word/Audio/LS", "Word/Audio/LM",
                 "Word/Audio/JL", "Word/Go/LS", "Word/Go/LM",
                 "Word/Go/JL"),
                ((-0.5, 0), (-1, 1), (-0.5, 1.5), (-0.5, 1.5), (-0.5, 1.5),
                 (-0.5, 1.5), (-0.5, 1.5), (-0.5, 1.5)),
                (
                        "base", "resp", "aud_ls", "aud_lm", "aud_jl", "go_ls",
                        "go_lm",
                        "go_jl")):
            times = [None, None]
            times[0] = t[0] - 0.5
            times[1] = t[1] + 0.5
            trials = trial_ieeg(good, epoch, times, preload=True)
            outliers_to_nan(trials, outliers=10)
            # spec = wavelet_scaleogram(trials, n_jobs=-2, decim=int(
            #     good.info['sfreq'] / 100))
            spec = cwt(trials, n_jobs=-2, decim=int(
                good.info['sfreq'] / 100))



from functools import singledispatch

import numpy as np
from mne import Epochs
from mne.epochs import BaseEpochs
from mne.time_frequency import AverageTFR, EpochsTFR
from mne.utils import logger, verbose

TFR = EpochsTFR | AverageTFR


def _log_rescale(baseline, mode='mean'):
    """Log the rescaling method."""
    if baseline is not None:
        msg = 'Applying baseline correction (mode: %s)' % mode
    else:
        msg = 'No baseline correction applied'
    return msg


@singledispatch
def rescale(data: np.ndarray, basedata: np.ndarray, mode: str = 'mean',
            copy: bool = False, axis: tuple[int] | int = -1) -> np.ndarray:
    """Rescale (baseline correct) data.

    Implement a variety of baseline correction methods. The data is
    modified in place by default.

    Parameters
    ----------
    data : array | mne.Epochs | mne.EpochsTFR
        It can be of any shape. The only constraint is that the last
        dimension should be time.
    basedata : array
        It can be of any shape. The last dimension should be time, and the
        other dimensions should be the same as data.
    mode : 'mean' | 'ratio' | 'logratio' | 'percent' | 'zscore' | 'zlogratio',\
        default 'mean', optional
        Perform baseline correction by
        - subtracting the mean of baseline values ('mean')
        - dividing by the mean of baseline values ('ratio')
        - dividing by the mean of baseline values and taking the log
        ('logratio')
        - subtracting the mean of baseline values followed by dividing by
        the mean of baseline values ('percent')
        - subtracting the mean of baseline values and dividing by the
        standard deviation of baseline values ('zscore')
        - dividing by the mean of baseline values, taking the log, and
        dividing by the standard deviation of log baseline values ('zlogratio')
    copy : bool, optional
        Whether to return a new instance or modify in place.
    axis : int or tuple[int], optional


    Returns
    -------
    data_scaled: array
        Array of same shape as data after rescaling.
    """
    if copy:
        data = data.copy()

    match mode:
        case 'mean':
            def fun(d, m, s):
                d -= m
        case 'ratio':
            def fun(d, m, s):
                d /= m
        case 'logratio':
            def fun(d, m, s):
                d /= m
                np.log10(d, out=d)
        case 'percent':
            def fun(d, m, s):
                d -= m
                d /= m
        case 'zscore':
            def fun(d, m, s):
                d -= m
                d /= s
        case 'zlogratio':
            def fun(d, m, s):
                d /= m
                np.log10(d, out=d)
                d /= s
        case _:
            raise NotImplementedError()
    mean = np.nanmean(basedata, axis=axis, keepdims=True)
    std = np.nanstd(basedata, axis=axis, keepdims=True)
    fun(data, mean, std)
    return data


@rescale.register
@verbose
def _(line: BaseEpochs, baseline: BaseEpochs,
      mode: str = 'mean', copy: bool = False, picks: list = 'data',
      verbose=None) -> Epochs:
    """Rescale (baseline correct) Epochs"""
    if copy:
        line: Epochs = line.copy()
    if verbose is not False:
        msg = _log_rescale(baseline, mode)
        logger.info(msg)

    # Average the baseline across epochs
    basedata = baseline.pick(picks)._data
    axes = list(range(basedata.ndim))

    # within channels
    axes.pop(1)

    # If time frequency then within frequency
    if isinstance(line, EpochsTFR):
        axes = (0, 3)
    elif isinstance(line, AverageTFR):
        axes = 2
    else:
        axes = tuple(axes)

    line.pick(picks)._data = rescale(line.pick(picks)._data, basedata, mode,
                                     False, axes)
    return line


@rescale.register
@verbose
def _(line: TFR, baseline: TFR,
      mode: str = 'mean', copy: bool = False, picks: list = 'data',
      verbose=None) -> Epochs:
    """Rescale (baseline correct) Epochs"""
    if copy:
        line: Epochs = line.copy()
    if verbose is not False:
        msg = _log_rescale(baseline, mode)
        logger.info(msg)

    # Average the baseline across epochs
    basedata = baseline.pick(picks)._data
    axes = list(range(basedata.ndim))

    # within channels
    axes.pop(1)

    # If time frequency then within frequency
    if isinstance(line, EpochsTFR):
        axes = (0, 3)
    elif isinstance(line, AverageTFR):
        axes = 2
    else:
        axes = tuple(axes)

    line.pick(picks)._data = rescale(line.pick(picks)._data, basedata, mode,
                                     False, axes)
    return line



from collections import Counter
from functools import cache, singledispatch
from typing import Union

import numpy as np
from mne import Epochs, event, events_from_annotations
from mne.epochs import BaseEpochs
from mne.io import Raw, base
from mne.time_frequency import AverageTFRArray, tfr_multitaper
from mne.utils import _pl, fill_doc, logger, verbose
from scipy import fft, signal, stats

from ieeg import ListNum
from ieeg.calc.scaling import rescale
from ieeg.calc.stats import sine_f_test
from ieeg.process import COLA, is_number
from ieeg.timefreq.utils import crop_pad, to_samples


class WindowingRemover(object):
    """Removes windowing artifacts from data.

    Parameters
    ----------
    sfreq : float
        The sampling frequency of the data.
    line_freqs : list of float
        The frequencies of the line noise.
    notch_width : list of float
        The notch widths for each line frequency.
    filter_length : int
        The length of the filter to use.
    low_bias : bool
        Whether to use a low bias filter.
    adaptive : bool
        Whether to use an adaptive filter.
    bandwidth : float
        The bandwidth of the multitaper windowing function.
    p_value : float
        The p-value to use in the F-test.
    verbose : bool
        Whether to print information.
    """

    @verbose
    def __init__(self, sfreq: float, line_freqs: ListNum,
                 notch_width: ListNum, filter_length: int, low_bias: bool,
                 adaptive: bool, bandwidth: float, p_value: float,
                 verbose: bool = None):
        self.sfreq = sfreq
        self.line_freqs = line_freqs
        self.notch_width = notch_width
        self.p_value = p_value
        self.filter_length = filter_length
        self.verbose = verbose
        self.low_bias = low_bias
        self.adaptive = adaptive
        self.bandwidth = bandwidth
        self.logger = logger
        self.rm_freqs = list()

    def dpss_windows(self, N: int, half_nbw: float, Kmax: int, *,
                     sym: bool = True, norm: Union[int, str] = None
                     ) -> tuple[np.ndarray, np.ndarray]:
        """Compute Discrete Prolate Spheroidal Sequences.

        Will give of orders [0,Kmax-1] for a given frequency-spacing multiple
        NW and sequence length N.

        .. note:: Copied from NiTime.

        Parameters
        ----------
        N : int
            Sequence length.
        half_nbw : float
            Standardized half bandwidth corresponding to 2 * half_bw = BW*f0
            = BW*N/dt but with dt taken as 1.
        Kmax : int
            Number of DPSS windows to return is Kmax (orders 0 through Kmax-1).
        sym : bool
            Whether to generate a symmetric window (`True`, for filter design)
            or a periodic window (`False`, for spectral analysis). Default is
            `True`.
        norm : 2 | 'approximate' | 'subsample' | None
            Window normalization method. If ``'approximate'`` or
            ``'subsample'``, windows are normalized by the maximum, and a
            correction scale-factor for even-length windows is applied either
            using ``N**2/(N**2+half_nbw)`` ("approximate") or a FFT-based
            subsample shift ("subsample"). ``2`` uses the L2 norm. ``None``
            (the default) uses ``"approximate"`` when ``Kmax=None`` and ``2``
            otherwise.

        Returns
        -------
        v, e : tuple,
            The v array contains DPSS windows shaped (Kmax, N).
            e are the eigenvalues.

        Notes
        -----
        Tridiagonal form of DPSS calculation (Slepian, 1978)

        References
        ----------
        David S. Slepian. Prolate spheroidal wave functions, fourier analysis,
        and uncertainty-V: the discrete case. Bell System Technical Journal,
        57(5):1371–1430, 1978. doi:10.1002/j.1538-7305.1978.tb02104.x.
        """

        dpss, eigvals = signal.windows.dpss(
            N, half_nbw, Kmax, sym=sym, norm=norm, return_ratios=True)
        if self.low_bias:
            idx = (eigvals > 0.9)
            if not idx.any():
                self.logger.warn('Could not properly use low_bias, keeping'
                                 'lowest-bias taper')
                idx = [np.argmax(eigvals)]
            dpss, eigvals = dpss[idx], eigvals[idx]
        assert len(dpss) > 0  # should never happen
        assert dpss.shape[1] == N  # old nitime bug
        return dpss, eigvals

    @fill_doc
    def params(self, n_times: int) -> tuple[np.ndarray, np.ndarray, bool]:
        """Triage windowing and multitaper parameters.

        Parameters
        ----------
        n_times : int
            The number of time points.

        Returns
        -------
        window_fun : array, shape=(n_tapers, n_times)
            The window functions for each taper.
        eigenvals : array, shape=(n_tapers,)
            The eigenvalues for each taper.
        adaptive : bool
            Whether to use adaptive weights to combine the tapered spectra into
            PSD
            """
        # Compute standardized half-bandwidth
        if isinstance(self.bandwidth, str):
            self.logger.info(
                '    Using standard spectrum estimation with "%s" window'
                % (self.bandwidth,))
            window_fun = signal.get_window(self.bandwidth, n_times)[np.newaxis]
            return window_fun, np.ones(1), False

        if self.bandwidth is not None:
            half_nbw = float(self.bandwidth) * n_times / self.sfreq
        else:
            half_nbw = 4.
        if half_nbw < 0.5:
            raise ValueError(
                'bandwidth value %s yields a normalized bandwidth of %s < 0.5,'
                ' use a value of at least %s'
                % (self.bandwidth, half_nbw, self.sfreq / n_times))

        # Compute DPSS windows
        n_tapers_max = int(np.floor(2 * half_nbw - 1))
        window_fun, eigvals = self.dpss_windows(
            n_times, half_nbw, n_tapers_max, sym=True)
        self.logger.info('  Using multitaper spectrum estimation with %d DPSS '
                         'windows' % len(eigvals))

        if self.adaptive and len(eigvals) < 3:
            self.logger.warn('Not adaptively combining the spectral estimators'
                             ' due to a low number of tapers (%s < 3).' % (
                                 len(eigvals),))
            self.adaptive = False

        return window_fun, eigvals, self.adaptive

    @cache
    def get_thresh(self, n_times: int = None) -> tuple[np.ndarray, float]:
        """Get the window function and threshold for given time points.

        Parameters
        ----------
        n_times : int | None
            The number of time points. If None, the filter length will be used.

        Returns
        -------
        window_fun : array, shape=(n_tapers, n_times)
            The window functions for each taper
        threshold : float
            The threshold for the F-statistic.
        """

        if n_times is None:
            n_times = self.filter_length

        # figure out what tapers to use
        window_fun, _, _ = self.params(n_times)

        # F-stat of 1-p point
        threshold = stats.f.ppf(1 - self.p_value / n_times, 2,
                                2 * len(window_fun) - 2)
        return window_fun, threshold

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """Remove line frequencies from data using multitaper method."""
        # Set default window function and threshold
        window_fun, thresh = self.get_thresh()
        n_times = x.shape[-1]
        n_samples = window_fun.shape[1]
        n_overlap = (n_samples + 1) // 2
        x_out = np.zeros_like(x)
        idx = [0]

        # Define how to process a chunk of data
        def process(x_):
            window_fun, thresh = self.get_thresh()
            out = _mt_remove(x_, self.sfreq, self.line_freqs, self.notch_width,
                             window_fun, thresh, self.get_thresh)
            self.rm_freqs.append(out[1])
            return (out[0],)  # must return a tuple

        # Define how to store a chunk of fully processed data (it's trivial)
        def store(x_):
            stop = idx[0] + x_.shape[-1]
            x_out[..., idx[0]:stop] += x_
            idx[0] = stop

        COLA(process, store, n_times, n_samples, n_overlap, self.sfreq,
             verbose=False).feed(x)
        assert idx[0] == n_times

        # report found frequencies, but do some sanitizing first by binning
        # into 1 Hz bins
        counts = Counter(sum((np.unique(np.round(ff)).tolist()
                              for f in self.rm_freqs for ff in f), list()))
        kind = 'Detected' if self.line_freqs is None else 'Removed'
        found_freqs = '\n'.join(f'    {freq:6.2f} : '
                                f'{counts[freq]:4d} window{_pl(counts[freq])}'
                                for freq in sorted(counts)) or '    None'
        self.logger.info(f'{kind} notch frequencies (Hz):\n{found_freqs}')

        x = x_out
        return x_out


def _mt_remove(x: np.ndarray, sfreq: float, line_freqs: ListNum,
               notch_widths: ListNum, window_fun: np.ndarray,
               threshold: float, get_thresh: callable,
               ) -> tuple[np.ndarray, list[float]]:
    """Use MT-spectrum to remove line frequencies.
    Based on Chronux. If line_freqs is specified, all freqs within notch_width
    of each line_freq is set to zero.
    """

    assert x.ndim == 1
    if x.shape[-1] != window_fun.shape[-1]:
        window_fun, threshold = get_thresh(x.shape[-1])
    # compute mt_spectrum (returning n_ch, n_tapers, n_freq)
    x_p, freqs = spectra(x[np.newaxis, :], window_fun, sfreq)
    f_stat, A = sine_f_test(window_fun, x_p)

    # find frequencies to remove
    indices = np.where(f_stat > threshold)[1]
    # pdf = 1-stats.f.cdf(f_stat, 2, window_fun.shape[0]-2)
    # indices = np.where(pdf < 1/x.shape[-1])[1]
    # specify frequencies within indicated ranges
    if line_freqs is not None and notch_widths is not None:
        if not isinstance(notch_widths, (list, tuple)) and is_number(
                notch_widths):
            notch_widths = [notch_widths] * len(line_freqs)
        ranges = [(freq - notch_width / 2, freq + notch_width / 2
                   ) for freq, notch_width in zip(line_freqs, notch_widths)]
        indices = [ind for ind in indices if any(
            lower <= freqs[ind] <= upper for (lower, upper) in ranges)]

    fits = list()
    # make "time" vector
    rads = 2 * np.pi * (np.arange(x.size) / float(sfreq))
    for ind in indices:
        c = 2 * A[0, ind]
        fit = np.abs(c) * np.cos(freqs[ind] * rads + np.angle(c))
        fits.append(fit)

    if len(fits) == 0:
        datafit = 0.0
    else:
        # fitted sinusoids are summed, and subtracted from data
        datafit = np.sum(fits, axis=0)

    return x - datafit, freqs[indices]


def spectra(x: np.ndarray, dpss: np.ndarray, sfreq: float,
            n_fft: int = None) -> tuple[np.ndarray, np.ndarray]:
    """Compute significant tapered spectra.

    Parameters
    ----------
    x : array, shape=(..., n_times)
        Input signal
    dpss : array, shape=(n_tapers, n_times)
        The tapers
    sfreq : float
        The sampling frequency
    n_fft : int | None
        Length of the FFT. If None, the number of samples in the input signal
        will be used.

    Returns
    -------
    x_mt : array, shape=(..., n_tapers, n_times)
        The tapered spectra
    freqs : array
        The frequency points in Hz of the spectra
    """
    if n_fft is None:
        n_fft = x.shape[-1]  # round(sfreq * round(x.shape[-1]*pad_fact/sfreq))

    # remove mean (do not use in-place subtraction as it may modify input x)
    x = x - np.mean(x, axis=-1, keepdims=True)

    # only keep positive frequencies
    freqs = fft.rfftfreq(n_fft, 1. / sfreq)

    # The following is equivalent to this, but uses less memory:
    x_mt = fft.rfft(x[:, np.newaxis, :] * dpss, n=n_fft, workers=1)
    # n_tapers = dpss.shape[0] if dpss.ndim > 1 else 1
    # x_mt = np.zeros(x.shape[:-1] + (n_tapers, len(freqs)),
    #                 dtype=np.complex128)
    # for idx, sig in enumerate(x):
    #     x_mt[idx] = fft.rfft(sig[..., np.newaxis, :] * dpss, n=n_fft)
    # Adjust DC and maybe Nyquist, depending on one-sided transform
    x_mt[..., 0] /= np.sqrt(2.)
    if n_fft % 2 == 0:
        x_mt[..., -1] /= np.sqrt(2.)
    return x_mt, freqs


@fill_doc
@singledispatch
@verbose
def spectrogram(line: BaseEpochs, freqs: np.ndarray,
                baseline: BaseEpochs = None, n_cycles: np.ndarray = None,
                pad: str = "0s", correction: str = 'ratio',
                verbose: int = None, **kwargs) -> AverageTFRArray:
    """Calculate the multitapered, baseline corrected spectrogram

    Parameters
    ----------
    line : Epochs
        The data to be processed
    %(freqs_tfr)s
    baseline : Epochs
        The baseline to be used for correction
    %(n_cycles_tfr)s
    pad : str
        The amount of padding to be removed in the spectrogram
    correction : str
        The type of baseline correction to be used
    %(verbose)s

    Notes
    -----
    %(time_bandwidth_tfr_notes)s

    Returns
    -------
    power : AverageTFR
        The multitapered, baseline corrected spectrogram
    """
    if n_cycles is None:
        n_cycles = freqs / 2

    power, itc = tfr_multitaper(line, freqs, n_cycles, verbose=verbose,
                                **kwargs)

    # crop the padding off the spectral estimates
    crop_pad(power, pad)

    if baseline is None:
        return power

    # apply baseline correction
    basepower, bitc = tfr_multitaper(baseline, freqs, n_cycles,
                                     verbose=verbose, **kwargs)
    crop_pad(basepower, pad)

    # set output data
    corrected_data = rescale(power._data, basepower._data, correction, axis=-1)

    return AverageTFRArray(power.info, corrected_data, power.times, freqs,
                           nave=power.nave, comment=power.comment,
                           method=power.method)


@spectrogram.register
def _(line: base.BaseRaw, freqs: np.ndarray, line_event: str, tmin: float,
      tmax: float, base_event: str = None, base_tmin: float = None,
      base_tmax: float = None, n_cycles: np.ndarray = None, pad: str = "500ms",
      correction: str = 'ratio', **kwargs) -> AverageTFRArray:
    """Calculate the multitapered, baseline corrected spectrogram

    Parameters
    ----------
    line : Raw
        The data to be processed
    freqs : array-like
        The frequencies to be used in the spectrogram
    line_event : str
        The event to be used for the spectrogram
    tmin : float
        The start time of the spectrogram
    tmax : float
        The end time of the spectrogram
    base_event : str
        The event to be used for the baseline
    base_tmin : float
        The start time of the baseline
    base_tmax : float
        The end time of the baseline
    n_cycles : array-like
        The number of cycles to be used in the spectrogram
    pad : str
        The amount of padding to be used in the spectrogram
    correction : str
        The type of baseline correction to be used

    Returns
    -------
    power : AverageTFR
        The multitapered, baseline corrected spectrogram
    """

    # determine the events
    events, ids = events_from_annotations(line)
    dat_ids = [ids[i] for i in event.match_event_names(ids, line_event)]

    # pad the data
    pad_secs = to_samples(pad, line.info['sfreq']) / line.info['sfreq']

    # Epoch the data
    data = Epochs(line, events, dat_ids, tmin - pad_secs,
                  tmax + pad_secs, baseline=None, preload=True)

    # run baseline corrected version
    if base_event is None:
        return spectrogram(data, freqs, None, n_cycles, pad, correction,
                           **kwargs)

    base_ids = [ids[i] for i in event.match_event_names(ids, base_event)]
    baseline = Epochs(line, events, base_ids, base_tmin - pad_secs,
                      base_tmax + pad_secs, baseline=None, preload=True)

    return spectrogram(data, freqs, baseline, n_cycles, pad, correction,
                       **kwargs)



""" Various methods for aligning microECoG datasets across patients.

Author: Zac Spalding
Cogan & Viventi Labs, Duke University
"""

from functools import reduce

import numpy as np
from sklearn.decomposition import PCA

from ieeg.decoding.joint_pca.utils import cnd_avg, label2str


class JointPCADecomp:
    def __init__(self, n_components=40, dim_red=PCA):
        """
        Initializes JointPCADecomp class with the number of latent
        components and the method for dimensionality reduction.

        Parameters
        ----------
        n_components : int, optional
            Number of components for dimensionality reduction i.e.
            dimensionality of latent space. Defaults to 40.
        dim_red : Callable, optional
            Dimensionality reduction function. Must implement
            sklearn-style fit_transform() function. Defaults to PCA.
        """

        self.n_components = n_components
        self.dim_red = dim_red

    def fit(self, X, y):
        """
        Learns source-specific (e.g. patient-specific) transformations to
        the shared latent space and stores transformations in self.transforms.

        Parameters
        ----------
        X : list of ndarray
            List of features from multiple sources to compute shared latent
            space.
        y : list of ndarray
            List of labels corresponding to feature sources. Must be the same
            length as features.
        """

        transforms = get_joint_PCA_transforms(X, y,
                                              n_components=self.n_components,
                                              dim_red=self.dim_red)
        self.transforms = transforms

    def transform(self, X, idx=-1):
        """
        Applies learned transformations to input data. Supports transforming
        a single, specified dataset or all-source datasets at once.

        Parameters
        ----------
        X : ndarray or list of ndarray
            Features to transform. If a list, the length must be equal to the
            number of learned transforms (i.e. transforming all sources).
            If an ndarray, a source-specific transformation is applied to the
            data, with the source specified by the idx input.
        idx : int, optional
            Index of saved transform list to apply to single source data,
            or -1 if applying to all sources. Defaults to -1.

        Raises
        ------
        IndexError
            Error if idx is too large to select a learned transform from the
            saved list.
        RuntimeError
            Error if fit() has not been called before calling transform().

        Returns
        -------
        ndarray or tuple
            Transformed data from single session if idx is not -1, or a tuple
            of containing:
                Transformed data from all sources input to the fit() method.
                Length will be equal to the number of learned transformations.
        """

        if not self._check_fit():
            raise RuntimeError('Must call fit() before transforming data.')
        if idx == -1:
            return self._transform_multiple(X)
        if idx >= len(self.transforms):
            raise IndexError('Input idx is greater than the number of learned '
                             'transforms. For transformation of data from a '
                             'specific session, provide the input idx as the '
                             'index of the session in the input list. If '
                             'transforming multiple sessions, set idx=-1 '
                             '(default).')
        return self._transform_single(X, idx)

    def fit_transform(self, X, y):
        """Fits the model with X and y and applies the learned transformations
        to X.

        Parameters
        ----------
        X : list of ndarray
            List of features from multiple sources to compute shared latent
            space.
        y : list of ndarray
            List of labels corresponding to feature sources. Must be the same
            length as features.

        Returns
        -------
        tuple
            Tuple containing transformed ndarray data from all sources input to
            the fit() method. Length will be equal to the number of learned
            transformations.
        """

        self.fit(X, y)
        return self.transform(X)

    def _transform_multiple(self, X):
        """
        Uses learned latent space transformations to transform data from all
        sources used to fit the decomposition.

        Parameters
        ----------
        X : list of ndarray
            List of features from multiple sources to compute shared latent
            space.

        Returns
        -------
        tuple of ndarray
            Tuple containing transformed ndarray data from all sources input to
             the fit() method. Length will be equal to the number of learned
             transformations.
        """

        transformed_lst = [0] * len(X)
        for i, (feats, transform) in enumerate(zip(X, self.transforms)):
            transform_feats = feats.reshape(-1, feats.shape[-1]) @ transform
            transformed_lst[i] = transform_feats.reshape(feats.shape[:-1] +
                                                         (-1,))
        return (*transformed_lst,)

    def _transform_single(self, X, idx):
        """
        Applies learned latent space transformations to data from a single
        source specified by idx.

        Parameters
        ----------
        X : ndarray
            Features to transform.
        idx : int
            Index of learned transforms to apply to X.

        Returns
        -------
        ndarray
            Features transformed to shared latent space.
        """

        transform = self.transforms[idx]
        transform_feats = X.reshape(-1, X.shape[-1]) @ transform
        return transform_feats.reshape(X.shape[:-1] + (-1,))

    def _check_fit(self):
        """Checks if the joint PCA decomposition has been fit to data.

        Returns
        -------
        boolean
            True if fit() has been called, False otherwise.
        """

        try:
            self.transforms
        except AttributeError:
            return False
        return True


class CCAAlign():

    def __init__(self, type='class', return_space='b_to_a'):
        self.type = type
        self.return_space = return_space

    def fit(self, X_a, X_b, y_a, y_b):
        L_a, L_b = reshape_latent_dynamics(X_a, X_b, y_a, y_b, type=self.type)
        M_a, M_b = CCA_align(L_a.T, L_b.T)
        self.M_a = M_a
        self.M_b = M_b

    def transform(self, X):
        if not self._check_fit():
            raise RuntimeError('Must call fit() before transforming data.')
        if self.return_space in ['b_to_a', 'a_to_b']:
            return self._transform_single(X)
        return self._transform_shared(X)

    def _transform_single(self, X):
        if self.return_space == 'b_to_a':
            return X @ self.M_b @ np.linalg.pinv(self.M_a)
        return X @ self.M_a @ np.linalg.pinv(self.M_b)

    def _transform_shared(self, X):
        return X[0] @ self.M_a, X[1] @ self.M_b

    def _check_fit(self):
        """Checks if the joint PCA decomposition has been fit to data.

        Returns
        -------
        boolean
            True if fit() has been called, False otherwise.
        """
        try:
            self.M_a
            self.M_b
        except AttributeError:
            return False
        return True


def get_joint_PCA_transforms(features, labels, n_components=40, dim_red=PCA):
    """
    Calculates a shared latent space across features from multiple patients
    or recording sessions.

    Uses the method described by Pandarinath et al. in
    https://www.nature.com/articles/s41592-018-0109-9 (2018) for pre-computing
    session specific read-in matrices (see Methods: Modifications to the LFADS
    algorithm for stitching together data from multiple recording sessions)

    Parameters
    ----------
    features : list
        List of features from multiple sources to compute shared latent space.
    labels : list
        List of labels corresponding to feature sources. Must be the same
        length as features.
    n_components : int, optional
        Number of components for dimensionality reduction i.e. dimensionality
        of latent space. Defaults to 40.
    dim_red : Callable, optional
        Dimensionality reduction function. Must implement sklearn-style
        fit_transform() function. Defaults to PCA.

    Returns
    -------
    tuple
        Tuple containing transformation matrices to shared latent space for
        each input source. Length will be equal to the length of the input
        feature list.
    """
    # process labels for easy comparison of label sequences
    labels = [label2str(labs) for labs in labels]

    # condition average firing rates for all datasets
    cnd_avg_data = [0] * len(features)
    for i, (feats, labs) in enumerate(zip(features, labels)):
        cnd_avg_data[i] = cnd_avg(feats, labs)

    # only use same conditions across datasets
    shared_lab = reduce(np.intersect1d, labels)
    cnd_avg_data = [cnd_avg_data[i][np.isin(np.unique(lab), shared_lab,
                                            assume_unique=True)] for i, lab
                    in enumerate(labels)]

    # combine all datasets into one matrix (n_conditions x n_timepoints x
    # sum channels)
    cross_pt_mat = np.concatenate(cnd_avg_data, axis=-1)
    # reshape to 2D with channels as final dim
    cross_pt_mat = cross_pt_mat.reshape(-1, cross_pt_mat.shape[-1])

    # perform dimensionality reduction on channel dim of combined matrix
    latent_mat = dim_red(n_components=n_components).fit_transform(cross_pt_mat)

    # calculate per pt channel -> factor transformation matrices
    pt_latent_trans = [0] * len(cnd_avg_data)
    for i, pt_ca in enumerate(cnd_avg_data):
        pt_ca = pt_ca.reshape(-1, pt_ca.shape[-1])  # isolate channel dim
        latent_trans = np.linalg.pinv(pt_ca) @ latent_mat  # lst_sq soln
        pt_latent_trans[i] = latent_trans
        # latent_trans = np.linalg.pinv(latent_mat) @ pt_ca  # lst_sq soln
        # pt_latent_trans[i] = latent_trans.T
        # pt_latent_trans[i] = np.linalg.pinv(latent_trans)

    return (*pt_latent_trans,)


def reshape_latent_dynamics(X_a, X_b, y_a, y_b, type='class'):
    if type == 'class':
        L_a, L_b = extract_latent_dynamics_by_class(X_a, X_b, y_a, y_b)
    elif type == 'trial':
        L_a, L_b = extract_latent_dynamics_by_trial_subselect(X_a, X_b, y_a,
                                                              y_b)
    else:
        raise ValueError('type must be "class" or "trial".')

    # fold timepoints into trials (isolate latent dimensionality as 2nd dim)
    L_a, L_b = L_a.reshape(-1, L_a.shape[-1]), L_b.reshape(-1, L_b.shape[-1])
    return L_a, L_b


def extract_latent_dynamics_by_class(X_a, X_b, y_a, y_b):
    # process labels for easy comparison of label sequences
    y_a, y_b = label2str(y_a), label2str(y_b)
    # average trials within class
    L_a, L_b = cnd_avg(X_a, y_a), cnd_avg(X_b, y_b)

    # only align via shared classes between datasets
    _, y_shared_a, y_shared_b = np.intersect1d(np.unique(y_a), np.unique(y_b),
                                               assume_unique=True,
                                               return_indices=True)
    L_a, L_b = L_a[y_shared_a], L_b[y_shared_b]

    return L_a, L_b


def extract_latent_dynamics_by_trial_subselect(X_a, X_b, y_a, y_b):
    y_a, y_b = label2str(y_a), label2str(y_b)
    L_a, L_b = shared_trial_subselect(X_a, X_b, y_a, y_b)
    return L_a, L_b


def shared_trial_subselect(X_a, X_b, y_a, y_b):
    L_a, L_b = [], []
    # subselect same amount of trials for each class
    for c in np.intersect1d(y_a, y_b):
        # shuffle trial order within class
        curr_a = np.random.permutation(np.where(y_a == c)[0])
        curr_b = np.random.permutation(np.where(y_b == c)[0])
        min_shared = min(curr_a.shape[0], curr_b.shape[0])

        L_a.append(X_a[curr_a[:min_shared]])
        L_b.append(X_b[curr_b[:min_shared]])
    L_a, L_b = np.vstack(L_a), np.vstack(L_b)
    return L_a, L_b


def CCA_align_by_class(X_a, X_b, y_a, y_b, return_space='b_to_a'):
    """
    CCA Alignment between 2 datasets with correspondence by averaging within
    class conditions.

    The number of features must be the same for datasets A and B. For example,
    if the datasets have different feature sizes, you can use PCA to reduce
    both datasets to the same number of PCs first.

    Parameters
    ----------
    X_a : ndarray
        Data matrix for dataset A of shape (n_trials_a, n_timepoints,
        n_features)
    X_b : ndarray
        Data matrix for dataset B of shape (n_trials_b, n_timepoints,
        n_features)
    y_a : ndarray
        Label matrix for dataset A of shape (n_trials_a, ...). The first
        dimension must be the trial dimension. This can be a 1D array, or a 2D
        array if each trial has multiple labels (e.g. a sequence of phonemes).
        Label sequences are converted to a single string so that only the same
        label sequences have correspondence between the datasets.
    y_b : ndarray
        Label matrix for dataset B of shape (n_trials_a, ...). See y_a for more
        details.
    return_space : str, optional
        How to perform alignment. Dataset B can be aligned to A, and vice versa
        ('b_to_a' and 'a_to_b', respectively), or both datasets can be aligned
        to a shared space ('shared'). Defaults to 'b_to_a'.

    Returns
    -------
    tuple
        Tuple containing aligned data matrix for dataset A and dataset B.
    """

    parse_return_type(return_space)

    # convert labels to strings for seqeunce comparison
    y_a = label2str(y_a)
    y_b = label2str(y_b)

    # group trials by label type
    L_a = cnd_avg(X_a, y_a)
    L_b = cnd_avg(X_b, y_b)

    # find common labels between datasets for alignment
    _, y_shared_a, y_shared_b = np.intersect1d(np.unique(y_a), np.unique(y_b),
                                               assume_unique=True,
                                               return_indices=True)
    L_a = L_a[y_shared_a]
    L_b = L_b[y_shared_b]

    # fold timepoints into trials
    L_a = np.reshape(L_a, (-1, L_a.shape[-1]))
    L_b = np.reshape(L_b, (-1, L_b.shape[-1]))

    # calculate manifold directions with CCA
    M_a, M_b = CCA_align(L_a.T, L_b.T)

    # align in put data with manifold transformation matrices
    if return_space == 'b_to_a':
        return X_a, X_b @ M_b @ np.linalg.pinv(M_a)
    elif return_space == 'a_to_b':
        return X_a @ M_a @ np.linalg.pinv(M_b), X_b
    return X_a @ M_a, X_b @ M_b


def CCA_align_by_trial_subselect(X_a, X_b, y_a, y_b, return_space='b_to_a'):
    """
    CCA Alignment between 2 datasets with correspondence via subselection of
    trials within shared clases.

    The number of features must be the same for datasets A and B. For example,
    if the datasets have different feature sizes, you can use PCA to reduce
    both datasets to the same number of PCs first.

    Parameters
    ----------
    X_a : ndarray
        Data matrix for dataset A of shape (n_trials_a, n_timepoints,
        n_features)
    X_b : ndarray
        Data matrix for dataset B of shape (n_trials_b, n_timepoints,
        n_features)
    y_a : ndarray
        Label matrix for dataset A of shape (n_trials_a, ...). The first
        dimension must be the trial dimension. This can be a 1D array, or a 2D
        array if each trial has multiple labels (e.g. a sequence of phonemes).
        Label sequences are converted to a single string so that only the same
        label sequences have correspondence between the datasets.
    y_b : ndarray
        Label matrix for dataset B of shape (n_trials_a, ...). See y_a for more
        details.
    return_space : str, optional
        How to perform alignment. Dataset B can be aligned to A, and vice versa
        ('b_to_a' and 'a_to_b', respectively), or both datasets can be aligned
        to a shared space ('shared'). Defaults to 'b_to_a'.

    Returns
    -------
    tuple
        Tuple containing aligned data matrix for dataset A and dataset B.
    """

    parse_return_type(return_space)

    y_a = label2str(y_a)
    y_b = label2str(y_b)

    L_a, L_b = [], []
    # subselect same amount of trials for each class
    for c in np.intersect1d(y_a, y_b):
        # shuffle trial order within class
        curr_a = np.random.permutation(np.where(y_a == c)[0])
        curr_b = np.random.permutation(np.where(y_b == c)[0])
        min_shared = min(curr_a.shape[0], curr_b.shape[0])

        L_a.append(X_a[curr_a[:min_shared]])
        L_b.append(X_b[curr_b[:min_shared]])

    # combine subselected trials
    L_a = np.vstack(L_a)
    L_b = np.vstack(L_b)

    # fold timepoints into trials
    L_a = np.reshape(L_a, (-1, L_a.shape[-1]))
    L_b = np.reshape(L_b, (-1, L_b.shape[-1]))

    # calculate alignment
    M_a, M_b = CCA_align(L_a.T, L_b.T)

    # align in put data with manifold transformation matrices
    if return_space == 'b_to_a':
        return X_a, X_b @ M_b @ np.linalg.pinv(M_a)
    elif return_space == 'a_to_b':
        return X_a @ M_a @ np.linalg.pinv(M_b), X_b
    return X_a @ M_a, X_b @ M_b


def parse_return_type(return_space):
    """
    Checks the CCA alignment return type is valid.

    Parameters
    ----------
    return_space : str
        String detailing how to perform alignment.

    Raises
    ------
    ValueError
        Error if return_space is not 'b_to_a', 'a_to_b', or 'shared'
    """

    if return_space not in ['b_to_a', 'a_to_b', 'shared']:
        raise ValueError('return_space must be "b_to_a" or "a_to_b" or'
                         '"shared".')


def CCA_align(L_a, L_b):
    """Canonical Correlation Analysis (CCA) alignment between 2 datasets.

    From: https://www.nature.com/articles/s41593-019-0555-4#Sec11.
    Returns manifold directions to transform L_a and L_b into a common space
    (e.g. L_a_new.T = L_a.T @ M_a, L_b_new.T = L_b.T @ M_b).
    To transform into a specific patient space, for example putting everything
    in patient A's space, use L_(b->a).T = L_b.T @ M_b @ (M_a)^-1, where L_a
    and L_(b->a) will be aligned in the same space.

    Parameters
    ----------
    L_a : ndarray
        Latent dynamics array for dataset A of shape (m, T),
        where m is the number of latent dimensions and T is the number of
        timepoints.
    L_b : ndarray
        Latent dynamics array for dataset B of shape (m, T)

    Returns
    -------
    tuple
        Tuple containing:
        M_a : ndarray
            Manifold directions for dataset A of shape (m, m)
        M_b : ndarray
            Manifold directions for dataset B of shape (m, m)

    Examples
    --------
    >>> import numpy as np
    >>> rand = np.random.RandomState(seed=0)
    >>> L_a = np.random.randn(5, 10)
    >>> L_b = np.random.randn(5, 10)
    >>> M_a, M_b = CCA_align(L_a, L_b)
    >>> M_a.shape
    (5, 5)
    >>> (L_b.T @ M_b).shape
    (10, 5)
    """
    # QR decomposition
    Q_a, R_a = np.linalg.qr(L_a.T)
    Q_b, R_b = np.linalg.qr(L_b.T)

    # SVD on q inner product
    U, S, Vt = np.linalg.svd(Q_a.T @ Q_b)

    # calculate manifold directions
    M_a = np.linalg.pinv(R_a) @ U
    M_b = np.linalg.pinv(R_b) @ Vt.T

    return M_a, M_b


if __name__ == '__main__':
    from timeit import timeit

    L_a = np.random.randn(50, 1000)
    L_b = np.random.randn(50, 1000)
    M_a, M_b = CCA_align(L_a, L_b)
    print(timeit('CCA_align(L_a, L_b)', globals=globals(), number=1000))



""" Alignment Util Fcns

Author: Zac Spalding
Cogan & Viventi Labs, Duke University
"""

import pickle

import numpy as np

from ieeg.calc.mat import Labels


def cnd_avg(data, labels):
    """Averages data trials along first axis by condition type present in
    labels.

    Args:
        data (ndarray): Data matrix with shape (n_trials, ...). The first
            dimension must be the trial dimension. Number and shape of other
            dimensions is arbitrary.
        labels (ndarray): Label array with shape (n_trials,).

    Returns:
        ndarray: Data matrix averaged within conditions with shape
        (n_conditions, ...).
    """
    data_shape = data.shape
    class_shape = (len(np.unique(labels)),) + data_shape[1:]
    data_by_class = np.zeros(class_shape)
    for i, seq in enumerate(np.unique(labels)):
        data_by_class[i] = np.mean(data[labels == seq], axis=0)
    return data_by_class


def label2str(labels):
    if not isinstance(labels, Labels):
        labels = Labels(labels, delim='')

    # Converts a 2D array of label sequences into a 1D array of label strings.
    if len(labels.shape) > 1:
        labels = labels.join(0)
    return labels.astype(str)


def save_pkl(data, filename):
    with open(filename, 'wb+') as f:
        pickle.dump(data, f, protocol=-1)


def load_pkl(filename):
    with open(filename, 'rb') as f:
        data = pickle.load(f)
    return data


def decoding_data_from_dict(data_dict, pt, p_ind, lab_type='phon',
                            algn_type='phon_seq'):
    D_tar, lab_tar, lab_tar_full = get_features_labels(data_dict[pt], p_ind,
                                                       lab_type, algn_type)

    pre_data = []
    for p_pt in data_dict[pt]['pre_pts']:
        D_curr, lab_curr, lab_curr_full = get_features_labels(data_dict[p_pt],
                                                              p_ind, lab_type,
                                                              algn_type)
        pre_data.append((D_curr, lab_curr, lab_curr_full))

    return (D_tar, lab_tar, lab_tar_full), pre_data


def get_features_labels(data, p_ind, lab_type, algn_type):
    lab_full = data['y_full_' + algn_type[:-4]]
    if p_ind == -1:  # collapsed across all phonemes
        D = data['X_collapsed']
        lab = data['y_' + lab_type + '_collapsed']
        lab_full = np.tile(lab_full, (3, 1))  # label repeat for shape match
    else:  # individual phoneme
        D = data['X' + str(p_ind)]
        lab = data['y' + str(p_ind)]
    if lab_type == 'artic':  # convert from phonemes to articulator label
        lab = phon_to_artic_seq(lab)
    return D, lab, lab_full


def phon_to_artic_seq(phon_seq):
    phon_to_artic_conv = {1: 1, 2: 1, 3: 2, 4: 2, 5: 3, 6: 3, 7: 3, 8: 4, 9: 4}
    flat_seq = phon_seq.flatten()
    artic_conv = np.array([phon_to_artic(phon_idx, phon_to_artic_conv) for
                           phon_idx in flat_seq])
    return np.reshape(artic_conv, phon_seq.shape)


def phon_to_artic(phon_idx, phon_to_artic_conv):
    return phon_to_artic_conv[phon_idx]
